# Copyright 2019 Kemal Kurniawan
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import deque
from datetime import timedelta
from typing import Any, Callable, Optional
from pathlib import Path
import logging
import pickle
import time


# TODO maybe remove this if EpochTimer works fine
def start_epoch_timer(*, key: str = '_epoch_start_time'):  # pragma: no cover
    """A callback factory to record the start time of an epoch.

    Together with `stop_epoch_timer`, this callback provides an alternative to
    `~rnnr.attachments.EpochTimer` when `~rnnr.attachments.ProgressBar` is used.

    Args:
        key: Store epoch start time in ``state[key]``.

    Returns:
        Callback that records starting time of epochs.
    """
    logger = logging.getLogger(f'{__name__}.epoch_timer')

    def callback(state):
        state[key] = time.time()
        if state['max_epoch'] > 1:
            logger.info('Starting epoch %d/%d', state['epoch'], state['max_epoch'])
        else:
            logger.info('Starting run')

    return callback


# TODO maybe remove this if EpochTimer works fine
def stop_epoch_timer(*, key: str = '_epoch_start_time'):  # pragma: no cover
    """A callback factory to report the elapsed time of an epoch.

    Together with `start_epoch_timer`, this callback provides an alternative to
    `~rnnr.attachments.EpochTimer` when `~rnnr.attachments.ProgressBar` is used.

    Args:
        key: Read epoch start time from ``state[key]``.

    Returns:
        Callback that reports the elapsed time of epochs.
    """
    logger = logging.getLogger(f'{__name__}.epoch_timer')

    def callback(state):
        elapsed = timedelta(seconds=time.time() - state[key])
        if state['max_epoch'] > 1:
            logger.info('Epoch %d/%d done in %s', state['epoch'], state['max_epoch'], elapsed)
        else:
            logger.info('Run is done in %s', elapsed)

    return callback


def maybe_stop_early(patience: int = 5, *, check: str = 'better', counter: str = '_counter'):
    """A callback factory for early stopping.

    The returned calback keeps a counter in ``state[counter]`` for the number of times
    ``state[check]`` is ``False``. If this counter exceeds ``patience``, the callback
    stops the runner by setting ``state['running'] = False``.

    Example:

        >>> valid_losses = [0.1, 0.2, 0.3]  # simulate validation batch losses
        >>> batches = range(10)
        >>>
        >>> from rnnr import Event, Runner
        >>> from rnnr.attachments import MeanReducer
        >>> from rnnr.callbacks import maybe_stop_early
        >>>
        >>> trainer = Runner()
        >>> @trainer.on(Event.EPOCH_STARTED)
        ... def print_epoch(state):
        ...     print('Epoch', state['epoch'], 'started')
        ...
        >>> @trainer.on(Event.EPOCH_FINISHED)
        ... def eval_on_valid(state):
        ...     def eval_fn(state):
        ...         state['output'] = state['batch']
        ...     evaluator = Runner()
        ...     evaluator.on(Event.BATCH, eval_fn)
        ...     MeanReducer(name='mean').attach_on(evaluator)
        ...     evaluator.run(valid_losses)
        ...     if state.get('best_loss', float('inf')) > evaluator.state['mean']:
        ...         state['better'] = True
        ...         state['best_loss'] = evaluator.state['mean']
        ...     else:
        ...         state['better'] = False
        ...
        >>> trainer.on(Event.EPOCH_FINISHED, maybe_stop_early(patience=2))
        >>> trainer.run(batches, max_epoch=7)
        Epoch 1 started
        Epoch 2 started
        Epoch 3 started
        Epoch 4 started

    Args:
        patience: Stop the runner when the counter exceeds this number.
        check: Increment counter if ``state[check]`` is ``False``.
        counter: Store the counter in ``state[counter]``.

    Returns:
        Callback that does early stopping.
    """
    logger = logging.getLogger(f'{__name__}.early_stopping')

    def callback(state):
        n = (state.get(counter, 0) + 1) if not state[check] else 0
        state[counter] = n
        if state[counter] > patience:
            logger.info('Patience exceeded, stopping early')
            state['running'] = False

    return callback


def checkpoint(
        what: str,
        *,
        under: Optional[Path] = None,
        at_most: int = 1,
        when: Optional[str] = None,
        using: Optional[Callable[[Any, Path], None]] = None,
        ext: str = 'pkl',
        prefix_fmt: str = '{epoch}_',
        queue_fmt: str = '_saved_{what}',
):
    """A callback factory for checkpointing.

    Checkpointing means saving some object stored in ``state[what]`` during a run under
    ``under`` directory with ``{prefix_fmt}{what}.{ext}`` as the filename.

    Example:

        >>> from pathlib import Path
        >>> from pprint import pprint
        >>> from rnnr import Event, Runner
        >>> from rnnr.callbacks import checkpoint
        >>>
        >>> batches = range(3)
        >>> tmp_dir = Path('/tmp')
        >>> runner = Runner()
        >>> @runner.on(Event.EPOCH_FINISHED)
        ... def store_checkpoint(state):
        ...     state['model'] = 'MODEL'
        ...     state['optimizer'] = 'OPTIMIZER'
        ...
        >>> runner.on(Event.EPOCH_FINISHED, checkpoint('model', under=tmp_dir, at_most=3))
        >>> runner.on(Event.EPOCH_FINISHED, checkpoint('optimizer', under=tmp_dir, at_most=3))
        >>> runner.run(batches, max_epoch=7)
        >>> pprint(sorted(list(tmp_dir.glob('*.pkl'))))
        [PosixPath('/tmp/5_model.pkl'),
         PosixPath('/tmp/5_optimizer.pkl'),
         PosixPath('/tmp/6_model.pkl'),
         PosixPath('/tmp/6_optimizer.pkl'),
         PosixPath('/tmp/7_model.pkl'),
         PosixPath('/tmp/7_optimizer.pkl')]

    Args:
        what: Get the object to save from ``state[what]``.
        under: Save the object under this directory. Defaults to the current working directory
            if not given.
        at_most: Maximum number of files saved. When the number of files exceeds this number,
            older files will be deleted.
        when: If given, only save the object when ``state[when]`` is ``True``.
        using: Function to invoke to save the object. If given, this must be a callable
            accepting two arguments: an object to save and a `Path` to save it to. The default
            is to save the object using `pickle`.
        ext: Extension for the filename.
        prefix_fmt: Format for the filename prefix. Any string keys in ``state`` can be used
            as replacement fields.
        queue_fmt: Keeps track of the saved files for the object with a queue stored in
            ``state[queue_fmt.format(what=what)]``.

    Returns:
        Callback that does checkpointing.
    """
    if under is None:  # pragma: no cover
        under = Path.cwd()
    if using is None:
        using = _save_with_pickle
    qkey = queue_fmt.format(what=what)
    logger = logging.getLogger(f'{__name__}.checkpointing')

    def callback(state):
        q = state.get(qkey, deque())
        if when is None or state[when]:
            fmt = f'{prefix_fmt}{what}.{ext}'
            path = under / fmt.format(**state)
            logger.info('Saving to %s', path)
            using(state[what], path)
            q.append(path)
        while len(q) > at_most:
            p = q.popleft()
            if p.exists():  # pragma: no cover
                p.unlink()
        state[qkey] = q

    return callback


def save(*args, **kwargs):  # pragma: no cover
    """An alias for `checkpoint`."""
    return checkpoint(*args, **kwargs)


def _save_with_pickle(obj: Any, path: Path) -> None:
    with open(path, 'wb') as f:
        pickle.dump(obj, f)



# Copyright 2019 Kemal Kurniawan
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from typing import Any, Callable, Dict, Iterable, List

from .event import Event

Callback = Callable[[dict], None]


class Runner:
    """A neural network runner.

    A runner provides a thin abstraction of iterating over batches for several epochs,
    which is typically done in neural network training. To customize the behavior during
    a run, a runner provides a way to listen to events emitted during such run.
    To listen to an event, call `Runner.on` and provide a callback which will be called
    when the event is emitted. An event callback is a callable that accepts a `dict`
    and returns nothing. The `dict` is the state of the run. By default, the state contains:

    * ``batches`` - Iterable of batches which constitutes an epoch.
    * ``max_epoch`` - Maximum number of epochs to run.
    * ``n_iters`` - Current number of batch iterations.
    * ``running`` - A boolean which equals ``True`` if the runner is still running. Can
      be set to ``False`` to stop the runner earlier.
    * ``epoch`` - Current number of epoch.
    * ``batch`` - Current batch retrieved from ``state['batches']``.

    Attributes:
        state (dict): Runner's state that is passed to event callbacks.

    Note:
        Callbacks for an event are called in the order they are passed to `~Runner.on`.

    Caution:
        All the state contents above are required for a runner to function properly.
        You are free to change their values to suit your use cases better, but be careful.
    """

    def __init__(self) -> None:
        self.state: dict = {}
        self._callbacks: Dict[Event, List[Callback]] = defaultdict(list)

    def on(self, event: Event, callbacks=None):
        """Add single/multiple callback(s) to listen to an event.

        If ``callbacks`` is ``None``, this method returns a decorator which accepts
        a single callback for the event. If ``callbacks`` is a sequence of callbacks,
        they will all be added as listeners to the event *in order*.

        Args:
            event: Event to listen.
            callbacks: Callback(s) for the event.

        Returns:
            A decorator which accepts a callback, if ``callbacks`` is ``None``.
        """
        if callbacks is not None:
            cblist = self._callbacks[event]
            try:
                cblist.extend(callbacks)
            except TypeError:  # must be a single callback
                cblist.append(callbacks)
            return

        def decorator(cb: Callback) -> Callback:
            self._callbacks[event].append(cb)
            return cb

        return decorator

    def run(self, batches: Iterable[Any], max_epoch: int = 1) -> None:
        """Run on batches for a number of epochs.

        Args:
            batches: Batches to iterate over in an epoch.
            max_epoch: Maximum number of epochs to run.
        """
        state = self.state
        state.update({
            'max_epoch': max_epoch,
            'batches': batches,
            'n_iters': 0,
            'running': True,
            'epoch': 0,
        })

        self._emit(Event.STARTED, state)

        while state['running'] and state['epoch'] < state['max_epoch']:
            state['epoch'] += 1
            self._emit(Event._ETIMER_STARTED, state)
            self._emit(Event.EPOCH_STARTED, state)
            self._emit(Event._REDUCER_RESET, state)
            self._emit(Event._PBAR_CREATED, state)

            state['_batches_iter'] = iter(state['batches'])
            self._run_epoch()

            self._emit(Event._PBAR_CLOSED, state)
            self._emit(Event._REDUCER_COMPUTED, state)
            self._emit(Event.EPOCH_FINISHED, state)
            self._emit(Event._ETIMER_FINISHED, state)

        self._emit(Event.FINISHED, state)
        state['running'] = False

    def resume(self, repeat_last_batch: bool = False) -> None:
        """Resume runner starting from the current state.

        Args:
            repeat_last_batch: Whether to repeat processing the last batch. Ignored if the
                last epoch is finished (i.e. the batches have been exhausted).
        """
        state = self.state
        state['running'] = True

        finished_last_epoch = state['n_iters'] % len(state['batches']) == 0

        if not finished_last_epoch:
            self._emit(Event._ETIMER_STARTED, state)
            self._emit(Event._PBAR_CREATED, state)

            if repeat_last_batch:
                self._emit(Event.BATCH, state)
                self._emit(Event._REDUCER_UPDATED, state)
                self._emit(Event._PBAR_UPDATED, state)

            self._run_epoch()

            self._emit(Event._PBAR_CLOSED, state)
            self._emit(Event._REDUCER_COMPUTED, state)
            self._emit(Event.EPOCH_FINISHED, state)
            self._emit(Event._ETIMER_FINISHED, state)

        while state['running'] and state['epoch'] < state['max_epoch']:
            state['epoch'] += 1
            self._emit(Event._ETIMER_STARTED, state)
            self._emit(Event.EPOCH_STARTED, state)
            self._emit(Event._REDUCER_RESET, state)
            self._emit(Event._PBAR_CREATED, state)

            state['_batches_iter'] = iter(state['batches'])
            self._run_epoch()

            self._emit(Event._PBAR_CLOSED, state)
            self._emit(Event._REDUCER_COMPUTED, state)
            self._emit(Event.EPOCH_FINISHED, state)
            self._emit(Event._ETIMER_FINISHED, state)

        self._emit(Event.FINISHED, state)
        state['running'] = False

    def _run_epoch(self) -> None:
        state = self.state
        while state['running']:
            try:
                state['batch'] = next(state['_batches_iter'])
            except StopIteration:
                break
            state['n_iters'] += 1
            self._emit(Event.BATCH, state)
            self._emit(Event._REDUCER_UPDATED, state)
            self._emit(Event._PBAR_UPDATED, state)

    def _emit(self, event: Event, state: dict) -> None:
        for callback in self._callbacks[event]:
            if not state['running']:
                break
            callback(state)



from unittest.mock import Mock, call
import pickle

from rnnr.callbacks import checkpoint


def test_ok(tmp_path):
    max_epoch, max_saved = 5, 2
    objs_values = {
        'model': [f'MODEL_{i}' for i in range(max_epoch)],
        'opt': [f'OPT_{i}' for i in range(max_epoch)],
    }

    ckpt_name = 'ckpt'
    callback = checkpoint(ckpt_name, under=tmp_path, at_most=max_saved)
    state = {}
    for i in range(max_epoch):
        ckpt = {name: values[i] for name, values in objs_values.items()}
        state.update({ckpt_name: ckpt, 'epoch': i + 1})
        callback(state)

    assert len(list(tmp_path.glob(f'*{ckpt_name}.pkl'))) == max_saved
    for i in range(max_saved):
        path = tmp_path / f'{max_epoch - i}_{ckpt_name}.pkl'
        assert path.exists()
        with open(path, 'rb') as f:
            assert pickle.load(f) == {
                name: values[max_epoch - i - 1]
                for name, values in objs_values.items()
            }


def test_conditional(tmp_path):
    max_epoch, max_saved = 5, 2
    objs_values = {
        'model': [f'MODEL_{i}' for i in range(max_epoch)],
        'opt': [f'OPT_{i}' for i in range(max_epoch)],
    }
    better_epochs = {1, 3, 5}

    ckpt_name = 'ckpt'
    callback = checkpoint(ckpt_name, under=tmp_path, at_most=max_saved, when='better')
    state = {}
    for i in range(max_epoch):
        ckpt = {name: values[i] for name, values in objs_values.items()}
        state.update({ckpt_name: ckpt, 'epoch': i + 1, 'better': i + 1 in better_epochs})
        callback(state)

    saved_epochs = [3, 5]  # the last 2
    assert {p.name for p in tmp_path.glob(f'*{ckpt_name}.pkl')} == \
        {f'{e}_{ckpt_name}.pkl' for e in saved_epochs}
    for e in saved_epochs:
        path = tmp_path / f'{e}_{ckpt_name}.pkl'
        assert path.exists()
        with open(path, 'rb') as f:
            assert pickle.load(f) == {
                name: values[e - 1]
                for name, values in objs_values.items()
            }


def test_save_fn(tmp_path):
    max_epoch = 5
    objs_values = {
        'model': [f'MODEL_{i}' for i in range(max_epoch)],
        'opt': [f'OPT_{i}' for i in range(max_epoch)],
    }
    mock_save_fn = Mock()

    ckpt_name = 'ckpt'
    callback = checkpoint(ckpt_name, under=tmp_path, using=mock_save_fn)
    for i in range(max_epoch):
        ckpt = {name: values[i] for name, values in objs_values.items()}
        callback({ckpt_name: ckpt, 'epoch': i + 1})

    assert mock_save_fn.mock_calls == [
        call({name: values[i]
              for name, values in objs_values.items()}, tmp_path / f'{i+1}_{ckpt_name}.pkl')
        for i in range(max_epoch)
    ]



import statistics as stat

import pytest

from rnnr import Event
from rnnr.attachments import MeanReducer


def test_ok(runner):
    batches = range(5)
    values = [12, 7, 8, 44, -13]

    @runner.on(Event.BATCH)
    def on_batch(state):
        state['output'] = values[state['batch']]

    r = MeanReducer('mean')
    r.attach_on(runner)
    runner.run(batches)

    assert r.name == 'mean'
    assert runner.state['mean'] == pytest.approx(stat.mean(values))


def test_more_than_one_epoch(runner):
    batches = range(5)
    values = [12, 7, 8, 44, -13, 78, 55, -109, 34, 10]

    @runner.on(Event.BATCH)
    def on_batch(state):
        ix = (state['epoch'] - 1) * len(batches) + state['batch']
        state['output'] = values[ix]

    def efcallback(state):
        assert state[r.name] == pytest.approx(
            stat.mean(values[(state['epoch'] - 1) * len(batches) + b] for b in batches))

    r = MeanReducer('mean')
    r.attach_on(runner)
    runner.on(Event.EPOCH_FINISHED, efcallback)
    runner.run(batches, max_epoch=2)


def test_value(runner):
    batches = range(10)

    @runner.on(Event.BATCH)
    def on_batch(state):
        state['value'] = state['batch']**3

    r = MeanReducer('mean', value='value')
    r.attach_on(runner)
    runner.run(batches)

    assert runner.state[r.name] == pytest.approx(stat.mean(b**3 for b in batches))


def test_size(runner):
    batches = range(5)
    sizes = [3, 4, 9, 10, 2]

    @runner.on(Event.BATCH)
    def on_batch(state):
        state['output'] = state['batch']
        state['foo'] = sizes[state['batch']]

    r = MeanReducer('mean', size='foo')
    r.attach_on(runner)
    runner.run(batches)

    assert runner.state[r.name] == pytest.approx(sum(batches) / sum(sizes))



from unittest.mock import Mock
import pickle

import pytest

from rnnr import Event, Runner


def test_init():
    r = Runner()
    assert len(r.state) == 0


def test_run(runner):
    batches, max_epoch = range(10), 5
    runner.run(batches, max_epoch=max_epoch)
    state = runner.state

    assert state['batches'] == batches
    assert state['max_epoch'] == max_epoch
    assert state['n_iters'] == len(batches) * max_epoch
    assert not state['running']


class TestOn:
    def test_started(self, runner):
        batches, max_epoch = range(10), 5

        def on_started(state):
            assert state['batches'] == batches
            assert state['max_epoch'] == max_epoch
            assert state['n_iters'] == 0
            assert state['running']
            assert state['epoch'] == 0

        runner.on(Event.STARTED, on_started)
        runner.run(batches, max_epoch=max_epoch)

    def test_epoch_started(self, runner):
        batches, max_epoch, n_calls = range(10), 5, 0

        def on_epoch_started(state):
            nonlocal n_calls
            assert state['batches'] == batches
            assert state['max_epoch'] == max_epoch
            assert state['epoch'] == n_calls + 1
            assert state['n_iters'] == n_calls * len(batches)
            assert state['running']
            n_calls += 1

        runner.on(Event.EPOCH_STARTED, on_epoch_started)
        runner.run(batches, max_epoch=max_epoch)

    def test_batch(self, runner):
        batches, max_epoch, n_calls = range(10), 5, 0

        @runner.on(Event.BATCH)
        def on_batch(state):
            nonlocal n_calls
            assert state['batches'] == batches
            assert state['max_epoch'] == max_epoch
            assert state['epoch'] == n_calls // len(batches) + 1
            assert state['batch'] == batches[n_calls % len(batches)]
            assert state['n_iters'] == n_calls + 1
            assert state['running']
            n_calls += 1

        runner.run(batches, max_epoch=max_epoch)
        assert n_calls == len(batches) * max_epoch

    def test_epoch_finished(self, runner):
        batches, max_epoch, n_calls = range(10), 5, 0

        def on_epoch_finished(state):
            nonlocal n_calls
            assert state['batches'] == batches
            assert state['max_epoch'] == max_epoch
            assert state['epoch'] == n_calls + 1
            assert state['n_iters'] == (n_calls + 1) * len(batches)
            assert state['running']
            n_calls += 1

        runner.on(Event.EPOCH_FINISHED, on_epoch_finished)
        runner.run(batches, max_epoch=max_epoch)

    def test_finished(self, runner):
        batches, max_epoch, n_calls = range(10), 5, 0

        def on_finished(state):
            assert state['batches'] == batches
            assert state['max_epoch'] == max_epoch
            assert state['n_iters'] == max_epoch * len(batches)
            assert state['running']
            nonlocal n_calls
            n_calls += 1

        runner.on(Event.FINISHED, on_finished)
        runner.run(batches, max_epoch=max_epoch)
        assert n_calls == 1

    def test_as_decorator(self, runner):
        n_calls, max_epoch = 0, 10

        @runner.on(Event.EPOCH_STARTED)
        def increment(state):
            nonlocal n_calls
            n_calls += 1

        runner.run(range(5), max_epoch=max_epoch)
        assert n_calls == max_epoch

    def test_multiple_callbacks(self, runner):
        mock_escb1, mock_escb2, max_epoch = Mock(), Mock(), 10
        runner.on(Event.EPOCH_STARTED, [mock_escb1, mock_escb2])
        runner.run(range(5), max_epoch=max_epoch)

        assert mock_escb1.call_count == max_epoch
        assert mock_escb2.call_count == max_epoch


class TestStop:
    def test_on_batch(self, runner):
        mock_escallback = Mock()
        mock_efcallback = Mock()
        n_calls = 0
        batches = range(10)

        def bcallback(state):
            nonlocal n_calls
            n_calls += 1
            if state['batch'] == 3:
                state['running'] = False

        runner.on(Event.EPOCH_STARTED, mock_escallback)
        runner.on(Event.BATCH, bcallback)
        runner.on(Event.EPOCH_FINISHED, mock_efcallback)
        runner.run(batches, max_epoch=2)

        assert mock_escallback.call_count == 1
        assert mock_efcallback.call_count == 0
        assert n_calls == 4
        assert runner.state['n_iters'] == 4
        assert runner.state['epoch'] == 1
        assert runner.state['batch'] == 3

    def test_on_epoch_started(self, runner):
        mock_efcallback = Mock()
        mock_bcallback = Mock()

        class MockBatches:
            n = 0

            def __iter__(self):
                self.n = 0
                return self

            def __next__(self):
                res = self.n
                self.n += 1
                if self.n >= 10:
                    raise StopIteration
                return res

        batches = MockBatches()

        def escallback(state):
            if state['epoch'] == 1:
                state['running'] = False

        runner.on(Event.EPOCH_STARTED, escallback)
        runner.on(Event.BATCH, mock_bcallback)
        runner.on(Event.EPOCH_FINISHED, mock_efcallback)
        runner.run(batches, max_epoch=7)

        assert mock_efcallback.call_count == 0
        assert mock_bcallback.call_count == 0
        assert batches.n == 0


class TestResume:
    def test_stopped_on_batch(self, tmp_path):
        from rnnr.attachments import ProgressBar
        batches, max_epoch = list(range(5)), 2
        bcallback_ncalls, efcallback_ncalls = 0, 0

        runner = Runner()
        ProgressBar().attach_on(runner)

        def bcallback(state):
            nonlocal bcallback_ncalls
            bcallback_ncalls += 1
            if state['stage'] == 'first' and state['n_iters'] == 2:
                state['running'] = False

        def efcallback(state):
            nonlocal efcallback_ncalls
            efcallback_ncalls += 1

        runner.on(Event.BATCH, bcallback)
        runner.on(Event.EPOCH_FINISHED, efcallback)
        runner.state['stage'] = 'first'
        runner.run(batches, max_epoch)
        with open(tmp_path / 'ckpt.pkl', 'wb') as f:
            pickle.dump(runner.state, f)

        with open(tmp_path / 'ckpt.pkl', 'rb') as f:
            ckpt = pickle.load(f)
        runner = Runner()
        ProgressBar().attach_on(runner)
        runner.on(Event.BATCH, bcallback)
        runner.on(Event.EPOCH_FINISHED, efcallback)
        runner.state.update(ckpt)
        runner.state['stage'] = 'second'
        runner.resume()

        assert bcallback_ncalls == len(batches) * max_epoch
        assert efcallback_ncalls == max_epoch

    def test_repeat_last_batch(self, tmp_path):
        from rnnr.attachments import LambdaReducer, MeanReducer, ProgressBar
        batches, max_epoch = list(range(5)), 2
        bcallback_ncalls, efcallback_ncalls = 0, 0

        runner = Runner()
        ProgressBar().attach_on(runner)
        LambdaReducer('total_output', lambda x, y: x + y).attach_on(runner)
        MeanReducer('mean_output').attach_on(runner)

        def bcallback(state):
            nonlocal bcallback_ncalls
            bcallback_ncalls += 1
            state['output'] = state['batch']
            if state['stage'] == 'first' and state['n_iters'] == 2:
                state['running'] = False

        def efcallback(state):
            nonlocal efcallback_ncalls
            efcallback_ncalls += 1
            assert state['total_output'] == sum(batches)
            assert state['mean_output'] == pytest.approx(sum(batches) / len(batches))

        runner.on(Event.BATCH, bcallback)
        runner.on(Event.EPOCH_FINISHED, efcallback)
        runner.state['stage'] = 'first'
        runner.run(batches, max_epoch)
        with open(tmp_path / 'ckpt.pkl', 'wb') as f:
            pickle.dump(runner.state, f)

        with open(tmp_path / 'ckpt.pkl', 'rb') as f:
            ckpt = pickle.load(f)
        runner = Runner()
        ProgressBar().attach_on(runner)
        LambdaReducer('total_output', lambda x, y: x + y).attach_on(runner)
        MeanReducer('mean_output').attach_on(runner)
        runner.on(Event.BATCH, bcallback)
        runner.on(Event.EPOCH_FINISHED, efcallback)
        runner.state.update(ckpt)
        runner.state['stage'] = 'second'
        runner.resume(repeat_last_batch=True)

        assert bcallback_ncalls == len(batches) * max_epoch + 1
        assert efcallback_ncalls == max_epoch

    def test_stopped_on_epoch(self, tmp_path):
        from rnnr.attachments import LambdaReducer, MeanReducer, ProgressBar
        batches, max_epoch = list(range(5)), 2
        bcallback_ncalls, efcallback_ncalls = 0, 0

        runner = Runner()
        ProgressBar().attach_on(runner)
        LambdaReducer('total_output', lambda x, y: x + y).attach_on(runner)
        MeanReducer('mean_output').attach_on(runner)

        def bcallback(state):
            nonlocal bcallback_ncalls
            bcallback_ncalls += 1
            state['output'] = state['batch']

        def efcallback(state):
            nonlocal efcallback_ncalls
            efcallback_ncalls += 1
            assert state['total_output'] == sum(batches)
            assert state['mean_output'] == pytest.approx(sum(batches) / len(batches))
            if state['stage'] == 'first' and state['epoch'] == 1:
                state['running'] = False

        runner.on(Event.BATCH, bcallback)
        runner.on(Event.EPOCH_FINISHED, efcallback)
        runner.state['stage'] = 'first'
        runner.run(batches, max_epoch)
        with open(tmp_path / 'ckpt.pkl', 'wb') as f:
            pickle.dump(runner.state, f)

        with open(tmp_path / 'ckpt.pkl', 'rb') as f:
            ckpt = pickle.load(f)
        runner = Runner()
        ProgressBar().attach_on(runner)
        LambdaReducer('total_output', lambda x, y: x + y).attach_on(runner)
        MeanReducer('mean_output').attach_on(runner)
        runner.on(Event.BATCH, bcallback)
        runner.on(Event.EPOCH_FINISHED, efcallback)
        runner.state.update(ckpt)
        runner.state['stage'] = 'second'
        runner.resume()

        assert bcallback_ncalls == len(batches) * max_epoch
        assert efcallback_ncalls == max_epoch



# Copyright 2019 Kemal Kurniawan
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import timedelta
from typing import Any, Callable, Optional, Type
from warnings import warn
import abc
import logging
import time

from tqdm import tqdm

from .event import Event
from .runner import Runner


class Attachment(abc.ABC):
    """An abstract base class for an attachment."""

    @abc.abstractmethod
    def attach_on(self, runner: Runner) -> None:
        """Attach to a runner.

        Args:
            runner: Runner to attach to.
        """
        pass


class EpochTimer(Attachment):  # pragma: no cover
    """An attachment to time epoch.

    Epochs are only timed when ``state['max_epoch']`` is greater than 1. At the start and
    end of every epoch, logging messages are written with log level of INFO.
    """
    logger = logging.getLogger(f'{__name__}.epoch_timer')
    _epoch_start_time = '_epoch_start_time'

    def attach_on(self, runner: Runner) -> None:
        runner.on(Event._ETIMER_STARTED, self._start)
        runner.on(Event._ETIMER_FINISHED, self._finish)

    def _start(self, state):
        if state['max_epoch'] > 1:
            if self._epoch_start_time not in state:
                state[self._epoch_start_time] = time.time()
                msg = 'Starting epoch %d/%d'
            else:
                msg = 'Resuming epoch %d/%d'
            self.logger.info(msg, state['epoch'], state['max_epoch'])

    def _finish(self, state):
        if state['max_epoch'] > 1:
            elapsed = timedelta(seconds=time.time() - state.pop(self._epoch_start_time))
            self.logger.info(
                'Epoch %d/%d done in %s', state['epoch'], state['max_epoch'], elapsed)


class ProgressBar(Attachment):
    """An attachment to display a progress bar.

    The progress bar is implemented using `tqdm`_.

    Example:

        >>> from rnnr import Runner
        >>> from rnnr.attachments import ProgressBar
        >>> runner = Runner()
        >>> ProgressBar().attach_on(runner)
        >>> runner.run(range(10), max_epoch=10)

    Args:
        n_items: Get the number of items in a batch from ``state[n_items]`` to update the
            progress bar with. If not given, the default is to always set it to 1.
        stats: Get the batch statistics from ``state[stats]`` and display it along
            with the progress bar. The statistics dictionary has the names of the statistics
            as keys and the statistics as values.
        **kwargs: Keyword arguments to be passed to `tqdm`_ class.


    .. _tqdm: https://github.com/tqdm/tqdm
    """
    _n_items_so_far = '_pbar_n_items_so_far'

    def __init__(
            self,
            *,
            n_items: str = 'n_items',
            stats: Optional[str] = None,
            tqdm_cls: Optional[Type[tqdm]] = None,
            **kwargs,
    ) -> None:
        if tqdm_cls is None:  # pragma: no cover
            tqdm_cls = tqdm

        self._tqdm_cls = tqdm_cls
        self._n_items = n_items
        self._stats = stats
        self._kwargs = kwargs

        self._pbar: tqdm

    def attach_on(self, runner: Runner) -> None:
        runner.on(Event._PBAR_CREATED, self._create)
        runner.on(Event._PBAR_UPDATED, self._update)
        runner.on(Event._PBAR_CLOSED, self._close)

    def _create(self, state: dict) -> None:
        n_items_so_far = state.get(self._n_items_so_far, 0)
        self._pbar = self._tqdm_cls(state['batches'], initial=n_items_so_far, **self._kwargs)
        state[self._n_items_so_far] = n_items_so_far

    def _update(self, state: dict) -> None:
        if self._stats is not None:
            self._pbar.set_postfix(**state[self._stats])
        n_items = state.get(self._n_items, 1)
        self._pbar.update(n_items)
        state[self._n_items_so_far] += n_items

    def _close(self, state: dict) -> None:
        self._pbar.close()
        state.pop(self._n_items_so_far)


class LambdaReducer(Attachment):
    """An attachment to compute a reduction over batches.

    This attachment gets the value of each batch and compute a reduction over them
    at the end of each epoch.

    Example:

        >>> from rnnr import Event, Runner
        >>> from rnnr.attachments import LambdaReducer
        >>> runner = Runner()
        >>> LambdaReducer('product', lambda x, y: x * y).attach_on(runner)
        >>> @runner.on(Event.BATCH)
        ... def on_batch(state):
        ...     state['output'] = state['batch']
        ...
        >>> runner.run([10, 20, 30])
        >>> runner.state['product']
        6000

    Args:
        name: Name of this attachment to be used as the key in the runner's
            state dict to store the reduction result.
        reduce_fn: Reduction function. It should accept two batch values and
            return their reduction result.
        value: Get the value of a batch from ``state[value]``.
    """

    def __init__(
            self,
            name: str,
            reduce_fn: Callable[[Any, Any], Any],
            *,
            value: str = 'output',
    ) -> None:
        self.name = name
        self._reduce_fn = reduce_fn
        self._value = value

    def attach_on(self, runner: Runner) -> None:
        runner.on(Event._REDUCER_RESET, self._reset)
        runner.on(Event._REDUCER_UPDATED, self._update)
        runner.on(Event._REDUCER_COMPUTED, self._compute)

    @property
    def _result(self) -> str:
        return f'_{self.name}_reducer_result'

    def _reset(self, state: dict) -> None:
        if self._result in state:  # pragma: no cover
            warn(
                f'You may have multiple reducers with name={self.name!r}, so one will '
                'overwrite the other.')
        state[self._result] = None

    def _update(self, state: dict) -> None:
        if state[self._result] is None:
            state[self._result] = state[self._value]
        else:
            state[self._result] = self._reduce_fn(state[self._result], state[self._value])

    def _compute(self, state: dict) -> None:
        state[self.name] = state.pop(self._result)


class MeanReducer(LambdaReducer):
    """An attachment to compute a mean over batch statistics.

    This attachment gets the value from each batch and compute their mean at the end of
    every epoch.

    Example:

        >>> from rnnr import Event, Runner
        >>> from rnnr.attachments import MeanReducer
        >>> runner = Runner()
        >>> MeanReducer('mean').attach_on(runner)
        >>> @runner.on(Event.BATCH)
        ... def on_batch(state):
        ...     state['output'] = state['batch']
        ...
        >>> runner.run([1, 2, 3])
        >>> runner.state['mean']
        2.0

    Args:
        name: Name of this attachment to be used as the key in the runner's state
            dict to store the mean value.
        value: Get the value of a batch from ``state[value]``.
        size: Get the size of a batch from ``state[size]``. If the state has no such key,
            the size defaults to 1. The sum of all these batch sizes is the divisor when
            computing the mean.
    """

    def __init__(
            self,
            name: str,
            *,
            value: str = 'output',
            size: str = 'size',
    ) -> None:
        super().__init__(name, lambda x, y: x + y, value=value)
        self._size = size

    @property
    def _total_size(self) -> str:
        return f'_{self.name}_reducer_total_size'

    def _reset(self, state: dict) -> None:
        super()._reset(state)
        state[self._total_size] = 0

    def _update(self, state: dict) -> None:
        super()._update(state)
        state[self._total_size] += state.get(self._size, 1)

    def _compute(self, state: dict) -> None:
        super()._compute(state)
        state[self.name] /= state.pop(self._total_size)


class SumReducer(LambdaReducer):  # pragma: no cover
    """An attachment to compute a sum over batch statistics.

    This attachment gets the value from each batch and compute their sum at the end of
    every epoch.

    Args:
        name: Name of this attachment to be used as the key in the runner's state
            dict to store the mean value.
        value: Get the value of a batch from ``state[value]``.
    """

    def __init__(self, name: str, *, value: str = 'output') -> None:
        super().__init__(name, lambda x, y: x + y, value=value)



from rnnr.callbacks import maybe_stop_early


def test_ok(runner):
    patience, counter, check, state = 7, 'cnt', 'check', {'running': True}
    callback = maybe_stop_early(patience, check=check, counter=counter)

    for i in range(patience):
        state[check] = False
        callback(state)
        assert state['running']
        assert state[counter] == i + 1

    state[check] = True
    callback(state)
    assert state['running']
    assert state[counter] == 0

    for i in range(patience + 1):
        state[check] = False
        callback(state)
    assert not state['running']
    assert state[counter] == patience + 1
